/*
 Infomaniak kDrive - iOS App
 Copyright (C) 2021 Infomaniak Network SA

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import UIKit
import InfomaniakLogin
import InfomaniakCore
import kDriveCore
import Lottie

class OnboardingViewController: UIViewController {

    @IBOutlet weak var navigationBar: UINavigationBar!
    @IBOutlet weak var collectionView: UICollectionView!
    @IBOutlet weak var pageControl: UIPageControl!
    @IBOutlet weak var nextButton: UIButton!
    @IBOutlet weak var signInButton: UIButton!
    @IBOutlet weak var registerButton: UIButton!
    @IBOutlet weak var buttonContentView: UIView!
    @IBOutlet weak var closeBarButtonItem: UIBarButtonItem!

    @IBOutlet weak var collectionViewTopConstraint: NSLayoutConstraint!
    @IBOutlet weak var signInButtonHeight: NSLayoutConstraint!
    @IBOutlet weak var nextButtonHeight: NSLayoutConstraint!
    @IBOutlet weak var registerButtonHeight: NSLayoutConstraint!

    var addUser = false
    var slides: [Slide] = []

    override func viewDidLoad() {
        super.viewDidLoad()
        InfomaniakLogin.setupWebviewNavbar(title: "", titleColor: nil, color: nil, buttonColor: nil, clearCookie: true, timeOutMessage: "Timeout")
        nextButton.setImage(KDriveAsset.arrowRight.image.withRenderingMode(.alwaysTemplate), for: .normal)
        nextButton.imageView?.tintColor = .white
        nextButton.accessibilityLabel = KDriveStrings.Localizable.buttonPlayerNext
        navigationBar.isHidden = !addUser
        navigationBar.isTranslucent = true
        navigationBar.setBackgroundImage(UIImage(), for: .default)
        navigationBar.shadowImage = UIImage()
        closeBarButtonItem.accessibilityLabel = KDriveStrings.Localizable.buttonClose
        collectionView.register(cellView: SlideCollectionViewCell.self)

        slides = createSlides()
        pageControl.numberOfPages = slides.count
        pageControl.currentPage = 0
        // Handle tap on control to change page
        pageControl.addTarget(self, action: #selector(pageChanged), for: .valueChanged)
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        signInButton.setLoading(false)
    }

    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        if view.frame.height < 600 {
            signInButtonHeight.constant = 45
            registerButtonHeight.constant = 45
            nextButtonHeight.constant = 60
            nextButton.layer.cornerRadius = nextButtonHeight.constant / 2
            collectionViewTopConstraint.constant = -40
        }
    }

    override func viewWillTransition(to size: CGSize, with coordinator: UIViewControllerTransitionCoordinator) {
        super.viewWillTransition(to: size, with: coordinator)
        coordinator.animate { _ in
            self.collectionView?.collectionViewLayout.invalidateLayout()
        } completion: { _ in
            let indexPath = IndexPath(row: self.pageControl.currentPage, section: 0)
            self.collectionView.scrollToItem(at: indexPath, at: .left, animated: true)
        }
    }

    @IBAction func nextButtonPressed(_ sender: Any) {
        if pageControl.currentPage < slides.count - 1 {
            let indexPath = IndexPath(row: pageControl.currentPage + 1, section: 0)
            collectionView.scrollToItem(at: indexPath, at: .left, animated: true)
        }
    }

    @objc func pageChanged() {
        if pageControl.currentPage < slides.count {
            let indexPath = IndexPath(row: pageControl.currentPage, section: 0)
            collectionView.scrollToItem(at: indexPath, at: .left, animated: true)
        }
    }

    @IBAction func signInButtonPressed(_ sender: Any) {
        InfomaniakLogin.webviewLoginFrom(viewController: self, delegate: self)
    }

    @IBAction func registerButtonPressed(_ sender: Any) {
        present(RegisterViewController.instantiateInNavigationController(delegate: self), animated: true)
    }

    @IBAction func closeButtonPressed(_ sender: Any) {
        dismiss(animated: true)
    }

    private func goToMainScreen() {
        UserDefaults.store(firstLaunch: false)
        UserDefaults.shared.numberOfConnections = 1
        (UIApplication.shared.delegate as! AppDelegate).setRootViewController(MainTabViewController.instantiate(), animated: true)
    }

    private func updateButtonsState() {
        if pageControl.currentPage == slides.count - 1 {
            if buttonContentView.isHidden {
                buttonContentView.alpha = 0
                buttonContentView.isHidden = false
                UIView.animate(withDuration: 0.25) {
                    self.buttonContentView.alpha = 1
                    self.buttonContentView.isHidden = false
                    self.nextButton.alpha = 0
                } completion: { _ in
                    self.nextButton.alpha = 0
                    self.nextButton.isHidden = true
                }
            }
        } else {
            if nextButton.isHidden {
                nextButton.alpha = 0
                nextButton.isHidden = false
                UIView.animate(withDuration: 0.25) {
                    self.nextButton.alpha = 1
                    self.nextButton.isHidden = false
                    self.buttonContentView.alpha = 0
                } completion: { _ in
                    self.buttonContentView.alpha = 0
                    self.buttonContentView.isHidden = true
                }
            }
        }
    }

    private func createSlides() -> [Slide] {
        let slide1 = Slide(backgroundImage: KDriveAsset.background1.image,
            illustrationImage: KDriveAsset.illuDevices.image,
            animationName: "illu_devices",
            title: KDriveStrings.Localizable.onBoardingTitle1,
            description: KDriveStrings.Localizable.onBoardingDescription1)

        let slide2 = Slide(backgroundImage: KDriveAsset.background2.image,
            illustrationImage: KDriveAsset.illuCollab.image,
            animationName: "illu_collab", title: KDriveStrings.Localizable.onBoardingTitle2,
            description: KDriveStrings.Localizable.onBoardingDescription2)

        let slide3 = Slide(backgroundImage: KDriveAsset.background3.image,
            illustrationImage: KDriveAsset.illuPhotos.image,
            animationName: "illu_photos",
            title: KDriveStrings.Localizable.onBoardingTitle3,
            description: KDriveStrings.Localizable.onBoardingDescription3)

        return [slide1, slide2, slide3]
    }

    class func instantiate() -> OnboardingViewController {
        return Storyboard.main.instantiateViewController(withIdentifier: "OnboardingViewController") as! OnboardingViewController
    }
}

// MARK: - UICollectionView Delegate

extension OnboardingViewController: UICollectionViewDelegate, UICollectionViewDataSource, UICollectionViewDelegateFlowLayout {

    func numberOfSections(in collectionView: UICollectionView) -> Int {
        return 1
    }

    func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
        return slides.count
    }

    func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        let cell = collectionView.dequeueReusableCell(type: SlideCollectionViewCell.self, for: indexPath)
        cell.isSmallDevice = view.frame.height < 600
        cell.configureCell(slide: slides[indexPath.row], isSmallDevice: view.frame.height < 600)
        return cell
    }

    func collectionView(_ collectionView: UICollectionView, willDisplay cell: UICollectionViewCell, forItemAt indexPath: IndexPath) {
        if let cell = cell as? SlideCollectionViewCell {
            cell.illustrationAnimationView.currentProgress = 0
            cell.illustrationAnimationView.play()
        }
    }

    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, insetForSectionAt section: Int) -> UIEdgeInsets {
        return .zero
    }

    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, sizeForItemAt indexPath: IndexPath) -> CGSize {
        return CGSize(width: collectionView.frame.width, height: collectionView.frame.height)
    }

    func scrollViewDidScroll(_ scrollView: UIScrollView) {
        let pageIndex = round(scrollView.contentOffset.x / view.frame.width)
        pageControl.currentPage = Int(pageIndex)
        updateButtonsState()
    }
}

// MARK: - Infomaniak Login Delegate
extension OnboardingViewController: InfomaniakLoginDelegate {

    func didCompleteLoginWith(code: String, verifier: String) {
        let previousAccount = AccountManager.instance.currentAccount
        signInButton.setLoading(true)
        registerButton.isEnabled = false
        AccountManager.instance.createAndSetCurrentAccount(code: code, codeVerifier: verifier) { account, error in
            if account != nil {
                // Download root file
                AccountManager.instance.currentDriveFileManager?.getFile(id: DriveFileManager.constants.rootID) { _, _, _ in
                    self.signInButton.setLoading(false)
                    self.registerButton.isEnabled = true
                    self.goToMainScreen()
                }
            } else {
                if previousAccount != nil {
                    AccountManager.instance.switchAccount(newAccount: previousAccount!)
                }
                DispatchQueue.main.async {
                    self.signInButton.setLoading(false)
                    self.registerButton.isEnabled = true
                    if let noDriveError = error as? InfomaniakCore.ApiError, noDriveError.code == DriveError.noDrive.code {
                        let driveErrorVC = DriveErrorViewController.instantiate()
                        driveErrorVC.driveErrorViewType = .noDrive
                        self.present(driveErrorVC, animated: true, completion: nil)
                    } else if let maintenanceError = error as? DriveError, maintenanceError.code == DriveError.maintenance.code {
                        let driveErrorVC = DriveErrorViewController.instantiate()
                        driveErrorVC.driveErrorViewType = .maintenance
                        self.present(driveErrorVC, animated: true, completion: nil)
                    } else {
                        self.okAlert(title: KDriveStrings.Localizable.errorTitle, message: KDriveStrings.Localizable.errorConnection, completion: nil)
                    }
                }
            }
        }
    }

    func didFailLoginWith(error: String) {
        signInButton.setLoading(false)
        self.registerButton.isEnabled = true
    }
}
